import json
import os
from contextlib import suppress
from types import TracebackType
from typing import Any


def obj_is_serializable(obj: Any) -> bool:
    """
    Determine if *obj* is serializable.

    :param obj: the reference object
    :return: *True* if serializable, *False* otherwise
    """
    # initialize the return variable
    result: bool = True

    # verify the object
    try:
        json.dumps(obj)
    except (TypeError, OverflowError):
        result = False

    return result


def obj_to_dict(obj: Any,
                omit_private: bool = True) -> dict[str, Any] | list[Any] | Any:
    """
    Convert the generic object *obj* to a *dict*.

    The conversion is done recursively. Attributes for which exceptions are raised on attempt
    to access them are silently omitted.

    :param obj: the object to be converted
    :param omit_private: whether to omit private attributes (defaults to *True*)
    :return: the dict obtained from *obj*
    """
    # declare the return variable
    result: dict[str, Any] | list[Any] | Any

    if isinstance(obj, dict):
        result = {str(k): obj_to_dict(obj=v,
                                      omit_private=omit_private) for k, v in obj.items()}
    elif isinstance(obj, list | tuple | set):
        result = [obj_to_dict(obj=item,
                              omit_private=omit_private) for item in obj]
    elif hasattr(obj, "__dict__") or not isinstance(obj, str | int | float | bool | type(None)):
        result = {}
        for attr in dir(obj):
            if not (omit_private and attr.startswith("_")):
                value: Any = getattr(obj,
                                     attr,
                                     None)
                if value is not None and not callable(value):
                    result[attr] = obj_to_dict(obj=value,
                                               omit_private=omit_private)
    else:
        result = obj

    return result


def obj_positional(o: Any,
                   /,
                   keys: tuple[Any, ...],
                   values: tuple[Any, ...],
                   def_value: Any = None) -> Any:
    """
    Locate the position of *o* within *keys*, and return the element in the same position in *values*.

    :param o: the source object
    :param keys: the tuple holding the keys to be inspected
    :param values: the tuple holding the positionally corresponding values
    :param def_value: the value to return, if not found (defaults to *None*)
    :return: the value positionally corresponding to the source object, or *def_value* if not found
    """
    # initialize the return variable
    result: Any = def_value

    with suppress(Exception):
        pos: int = keys.index(o)
        result = values[pos]

    return result


def exc_format(exc: Exception,
               exc_info: tuple[type[BaseException], BaseException, TracebackType]) -> str:
    """
    Format the error message resulting from the exception raised in execution time.

    The format to use: <python_module>, <line_number>: <exc_class> - <exc_text>

    :param exc: the exception raised
    :param exc_info: information associated with the exception
    :return: the formatted message
    """
    tback: TracebackType = exc_info[2]
    cls: str = str(exc.__class__)

    # retrieve the execution point where the exception was raised (bottom of the stack)
    tlast: TracebackType = tback
    while tlast.tb_next:
        tlast = tlast.tb_next

    # retrieve the module name and the line number within the module
    try:
        fname: str = os.path.split(p=tlast.tb_frame.f_code.co_filename)[1]
    except Exception:  # noqa: # noinspection PyBroadException
        fname: str = "<unknow module>"
    fline: int = tlast.tb_lineno

    return f"{fname}, {fline}, {cls[8:-2]} - {exc}"



import ast
import os
from base64 import b64decode, urlsafe_b64decode
from contextlib import suppress
from datetime import date
from dateutil import parser
from dateutil.parser import ParserError
from enum import IntEnum, StrEnum
from pathlib import Path
from typing import Any, Final, Literal

# the prefix for the names of the environment variables
APP_PREFIX: Final[str] = os.getenv(key="PYPOMES_APP_PREFIX",
                                   default="")


def env_get_str(key: str,
                values: list[str] = None,
                ignore_case: bool = False,
                def_value: str = None) -> str | None:
    """
    Retrieve the string value defined for *key* in the current operating environment.

    If *values* is specified, the value obtained is checked for occurrence therein.

    :param key: the key which the value is associated with
    :param values: optional list of valid values
    :param ignore_case: specifies whether to ignore capitalization
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the string value associated with the key, or *def_value* if error
    """
    # initialize the return variable
    result: str | None = None

    value: str = os.getenv(key)
    # allow for value to be defined as an empty string
    if value is None:
        result = def_value
    elif values:
        val: str = value
        if ignore_case:
            val = value.lower() if isinstance(value, str) else value
            values = [v.lower() if isinstance(v, str) else v for v in values]
        if val in values:
            result = value
    else:
        result = value

    return result


def env_get_strs(key: str,
                 values: list[str] = None,
                 ignore_case: bool = False) -> list[str] | None:
    """
    Retrieve the string values defined for *key* in the current operating environment.

    The values must be provided as a comma-separated list of strings.
    If *values* is specified, the values obtained are checked for occurrence therein, with *ignore_case*
    determining whether to ignore capitalization. On failure, *None* is returned.

    :param key: the key the values ares associated with
    :param values: optional list of valid values
    :param ignore_case: specifies whether to ignore capitalization when checking with *values*
    :return: the string values associated with the key, or *None* if error or no values found
    """
    # initialize the return variable
    result: list[str] | None = None

    vals: str = os.getenv(key)
    if vals:
        result = vals.split(",")
        if values:
            if ignore_case:
                vals = vals.lower()
                values = [v.lower() if isinstance(v, str) else v for v in values]
            for val in vals.split(","):
                if val not in values:
                    result = None
                    break
    return result


def env_get_int(key: str,
                values: list[int] = None,
                def_value: int = None) -> int | None:
    """
    Retrieve the integer value defined for *key* in the current operating environment.

    If *values* is specified, the value obtained is checked for occurrence therein.

    :param key: the key which the value is associated with
    :param values: optional list of valid values
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the integer value associated with the key, or *def_value* if error
    """
    result: int | None
    try:
        result = int(os.environ[key])
        if values and result not in values:
            result = None
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_get_ints(key: str,
                 values: list[str] = None) -> list[int] | None:
    """
    Retrieve the integer values defined for *key* in the current operating environment.

    The values must be provided as a comma-separated list of integers.
    If *values* is specified, the values obtained are checked for occurrence therein.
    On failure, *None* is returned.

    :param key: the key the values ares associated with
    :param values: optional list of valid values
    :return: the integer values associated with the key, or *None* if error or no values found
    """
    result: list[int] | None = None
    # noinspection PyUnusedLocal
    with suppress(Exception):
        vals: list[str] = os.environ[key].split(",")
        if vals:
            result = [int(val) for val in vals]
            if values:
                for val in result:
                    if val not in values:
                        result = None
                        break
    return result


def env_get_float(key: str,
                  values: list[float] = None,
                  def_value: float = None) -> float | None:
    """
    Retrieve the float value defined for *key* in the current operating environment.

    If *values* is specified, the value obtained is checked for occurrence therein.

    :param key: the key which the value is associated with
    :param values: optional list of valid values
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the float value associated with the key, or *def_value* if error
    """
    result: float | None
    try:
        result = float(os.environ[key])
        if values and result not in values:
            result = None
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_get_floats(key: str,
                   values: list[str] = None) -> list[float] | None:
    """
    Retrieve the float values defined for *key* in the current operating environment.

    The values must be provided as a comma-separated list of floats.
    If *values* is specified, the values obtained are checked for occurrence therein.
    On failure, *None* is returned.

    :param key: the key the values ares associated with
    :param values: optional list of valid values
    :return: the float values associated with the key, or *None* if error or no values found
    """
    result: list[float] | None = None
    # noinspection PyUnusedLocal
    with suppress(Exception):
        vals: list[str] = os.environ[key].split(",")
        if vals:
            result = [float(val) for val in vals]
            if values:
                for val in result:
                    if val not in values:
                        result = None
                        break
    return result


def env_get_enum(key: str,
                 enum_class: type[IntEnum | StrEnum],
                 values: list[IntEnum | StrEnum] = None,
                 def_value: IntEnum | StrEnum = None) -> Any:
    """
    Retrieve the enum value defined for *key* in the current operating environment.

    If provided, this value must be a name or a value corresponding to an instance of a subclass of *enum_class*.
    The only accepted values for *enum_class* are subclasses of  *StrEnum* or *IntEnum*.
    If *values* is specified, the value obtained is checked for occurrence therein.
    On failure, *None* is returned.

    :param key: the key which the value is associated with
    :param enum_class: the *enum* class to consider (must be a subclass of *IntEnum* or *StrEnum*)
    :param values: optional list of allowed values (defaults to all elements of *enum_class*)
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the value associated with the key, or *def_value* if error
    """
    from .enum_pomes import IntEnumUseName

    # initialize the return variable
    result: Any = None

    if issubclass(enum_class, IntEnumUseName):
        # noinspection PyUnresolvedReferences
        name: str = env_get_str(key=key,
                                values=[e.name for e in (values or enum_class)],
                                ignore_case=True,
                                def_value=def_value.name if def_value else None)
        if isinstance(name, str):
            for e in enum_class:
                if e.name.lower() == name.lower():
                    result = e
                    break
    else:
        value: Any = None
        vals: list = [e.value for e in (values or enum_class)]
        if issubclass(enum_class, StrEnum):
            value: str = env_get_str(key=key,
                                     values=vals,
                                     def_value=def_value)
        elif issubclass(enum_class, IntEnum):
            value: int = env_get_int(key=key,
                                     values=vals,
                                     def_value=def_value)
        if value:
            result = enum_class(value)

    return result


def env_get_enums(key: str,
                  enum_class: type[IntEnum | StrEnum]) -> list:
    """
    Retrieve the enum values defined for *key* in the current operating environment.

    If provided, these value must be names or values corresponding to an instance of a subclass of *enum_class*.
    The only accepted values for *enum_class* are subclasses of *StrEnum* or *IntEnum*. It must be possible
    to convert all names associated with *key* to enum instances. On failure, *None* is returned.

    :param key: the key which the values are associated with
    :param enum_class: the *enum* class to consider (must be a subclass of *IntEnum* or *StrEnum*)
    :return: the values associated with the key, or *None* if error or no values found
    """
    from .enum_pomes import IntEnumUseName

    # initialize the return variable
    result: list | None = None

    enums: list = []
    values: str = os.getenv(key)
    if values:
        found: bool = False
        names: list[str] = values.split(",")
        for name in names:
            found = False
            if issubclass(enum_class, IntEnumUseName):
                for e in enum_class:
                    if e.name.lower() == name.lower():
                        enums.append(e)
                        found = True
                        break
            else:
                if issubclass(enum_class, IntEnum) and name.isdigit():
                    value: int = int(name)
                else:
                    value: str = name
                if value in enum_class:
                    # noinspection PyCallingNonCallable
                    enums.append(enum_class(value))
                    found = True
            # break on the first failure
            if not found:
                break
        if found:
            result = enums

    return result


def env_get_bool(key: str,
                 def_value: bool = None) -> bool:
    """
    Retrieve the boolean value defined for *key* in the current operating environment.

    These are the criteria:
        - case is disregarded
        - the string values accepted to stand for *True* are *1*, *t*, or *true*
        - the string values accepted to stand for *False* are *0*, *f*, or *false*
        - all other values causes *None* to be returned

    :param key: the key which the value is associated with
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the boolean value associated with the key, or *def_value* if error
    """
    result: bool | None
    try:
        if os.environ[key].lower() in ["1", "t", "true"]:
            result = True
        elif os.environ[key].lower() in ["0", "f", "false"]:
            result = False
        else:
            result = None
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_get_bytes(key: str,
                  encoding: Literal["base64", "base64url", "hex", "utf8"] = "base64url",
                  values: list[bytes] = None,
                  def_value: bytes = None) -> bytes | None:
    """
    Retrieve the byte value defined for *key* in the current operating environment.

    The string defined in the environment must contain the *bytes* value encoded as defined in *encoding*.
    If *values* is specified, the value obtained is checked for occurrence therein.
    On failure, *None* is returned.

    :param key: the key which the value is associated with
    :param encoding: the representation of the *bytes* value
    :param values: optional list of valid values
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the byte value associated with the key, or *def_value* if error
    """
    result: bytes | None = None
    try:
        value: str = os.environ[key]
        match encoding:
            case "hex":
                result = bytes.fromhex(value)
            case "utf8":
                result = value.encode()
            case "base64":
                result = b64decode(s=value)
            case "base64url":
                result = urlsafe_b64decode(s=value)
        if values and result not in values:
            result = None
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_get_date(key: str,
                 def_value: date = None) -> date:
    """
    Retrieve the date value defined for *key* in the current operating environment.

    :param key: the key which the value is associated with
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the date value associated with the key, or *def_value* if error
    """
    result: date
    try:
        result = parser.parse(os.environ[key]).date()
    except (AttributeError, KeyError, TypeError, ParserError, OverflowError):
        result = def_value

    return result


def env_get_path(key: str,
                 def_value: Path = None) -> Path:
    """
    Retrieve the path value defined for *key* in the current operating environment.

    :param key: the key which the value is associated with
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the path value associated with the key, or *def_value* if error
    """
    result: Path
    try:
        result = Path(os.environ[key]).resolve()
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_get_obj(key: str,
                def_value: Any = None) -> Any:
    """
    Retrieve the string-marshalled object defined for *key* in the current operating environment.

    As an example, suppose the string *"{'name': 'Alice', 'age': 30, 'city': 'New York'}"*.
    When properly unmarshalled, it yields the *dict* object
        {
            'name': 'Alice',
            'age': 30,
            'city': 'New York'
        }

    :param key: the key which the value is associated with
    :param def_value: the value to return, if obtaining the value for *key* fails (defaults to *None*)
    :return: the unmarshalled object associated with the key, or *def_value* if error
    """
    result: Any
    try:
        result = ast.literal_eval(node_or_string=os.environ[key])
    except (AttributeError, KeyError, TypeError):
        result = def_value

    return result


def env_is_docker() -> bool:
    """
    Determine whether the application is running inside a Docker container.

    Note that a reasonable, but not infallible, heuristics is used.

    :return: 'True' if this could be determined, 'False' otherwise
    """
    result: bool = Path("/.dockerenv").exists()
    if not result:
        with (suppress(Exception),
              Path("/proc/1/cgroup", "rt").open() as f):
            result = "docker" in f.read()

    return result



import sys
from email.message import EmailMessage
from enum import StrEnum, auto
from logging import Logger
from smtplib import SMTP, SMTP_SSL
from typing import Any

from .file_pomes import Mimetype
from .env_pomes import APP_PREFIX, env_get_int, env_get_str


class EmailParam(StrEnum):
    """
    Parameters for email.
    """
    HOST = auto()
    PORT = auto()
    ACCOUNT = auto()
    PWD = auto()
    DEFAULT_FROM = auto()
    SECURITY = auto()


_EMAIL_CONFIG: dict[EmailParam, Any] = {
    EmailParam.HOST: env_get_str(key=f"{APP_PREFIX}_EMAIL_HOST"),
    EmailParam.PORT: env_get_int(key=f"{APP_PREFIX}_EMAIL_PORT"),
    EmailParam.ACCOUNT: env_get_str(key=f"{APP_PREFIX}_EMAIL_ACCOUNT"),
    EmailParam.PWD: env_get_str(key=f"{APP_PREFIX}_EMAIL_PWD"),
    EmailParam.DEFAULT_FROM: env_get_str(key=f"{APP_PREFIX}_EMAIL_DEFAULT_FROM"),
    EmailParam.SECURITY: env_get_str(key=f"{APP_PREFIX}_EMAIL_SECURITY")
}


def email_setup(host: str,
                port: int,
                account: str,
                pwd: str,
                origin: str,
                security: str = None) -> None:
    """
    Configure the email server.

    Invoking this function overrides the configuration parameters obtained from environment variables.

    :param host: the host URL
    :param port: the connection port (a positive integer)
    :param account: the logon account
    :param pwd: the logon password
    :param origin: the address of origin for the e-mails
    :param security: the security protocol ('ssl' and 'tls' are currently supported)
    """
    global _EMAIL_CONFIG
    _EMAIL_CONFIG = {
        EmailParam.HOST: host,
        EmailParam.PORT: port,
        EmailParam.ACCOUNT: account,
        EmailParam.PWD: pwd,
        EmailParam.DEFAULT_FROM: origin or account,
        EmailParam.SECURITY: security
    }


def email_send(email_to: str,
               subject: str,
               content: str,
               mimetype: Mimetype = Mimetype.TEXT,
               email_from: str = None,
               errors: list[str] = None,
               logger: Logger = None) -> None:
    """
    Send email to *user_email*, with *subject* as the email subject, and *content* as the email message.

    :param email_to: the address to send the email to
    :param subject: the email subject
    :param content: the email message
    :param mimetype: the mimetype of the content (defaults to *text/plain*)
    :param email_from: the email address of origin (defaults to the configured origin)
    :param errors: incidental error messages
    :param logger: optional logger
    """
    # import needed function
    from .obj_pomes import exc_format

    # build the email object
    email_msg = EmailMessage()
    email_msg["From"] = (email_from or
                         _EMAIL_CONFIG[EmailParam.DEFAULT_FROM] or
                         _EMAIL_CONFIG[EmailParam.ACCOUNT])
    email_msg["To"] = email_to
    email_msg["Subject"] = subject
    maintype, subtype = mimetype.split("/")
    # BUG HANDLING:
    #   will crash if parameter 'maintype' is passed and 'content' is a string
    if isinstance(content, str):
        email_msg.set_content(content,
                              subtype=subtype)
    else:
        email_msg.set_content(content,
                              maintype=maintype,
                              subtype=subtype)
    # send the message
    try:
        # instantiate the email server, login and send the email
        if _EMAIL_CONFIG[EmailParam.SECURITY] == "ssl":
            with SMTP_SSL(host=_EMAIL_CONFIG[EmailParam.HOST],
                          port=_EMAIL_CONFIG[EmailParam.PORT]) as server:
                server.login(user=_EMAIL_CONFIG[EmailParam.ACCOUNT],
                             password=_EMAIL_CONFIG[EmailParam.PWD])
                server.send_message(msg=email_msg)
        else:
            with SMTP(host=_EMAIL_CONFIG[EmailParam.HOST],
                      port=_EMAIL_CONFIG[EmailParam.PORT]) as server:
                if _EMAIL_CONFIG[EmailParam.SECURITY] == "tls":
                    server.starttls()

                # possible exceptions:
                #   - SMTPAuthenticationError: the server didn't accept the username/password combination
                #   - SMTPException: no suitable authentication method was found
                #   - SMTPHeloError: the server didn't reply properly to the helo greeting
                #   - SMTPNotSupportedError: the AUTH command is not supported by the server
                #   - SMTPServerDisconnected: the connection was unexpectedly closed
                server.login(user=_EMAIL_CONFIG[EmailParam.ACCOUNT],
                             password=_EMAIL_CONFIG[EmailParam.PWD])
                server.send_message(msg=email_msg)
        if logger:
            logger.debug(msg=f"Sent email '{subject}' to '{email_to}'")
    except Exception as e:
        # the operation raised an exception
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        err_msg: str = f"Error sending the email: {exc_err}"
        if logger:
            logger.error(msg=err_msg)
        if isinstance(errors, list):
            errors.append(err_msg)


def email_codify(email: str) -> str:
    """
    Codify *email* to provide a hint at its content, whilst preventing its usage.

    The codification process changes my_mail@my_server.com into m*****l@m********.com.

    :param email: the email to codify
    :return: the codified email
    """
    # initialize the return variable
    result: str = email

    pos1: int = email.rfind("@")
    pos2: int = email.rfind(".")
    if pos2 > pos1 > 0:
        result = email[0] + "*" * (pos1 - 2) + \
                 email[pos1 - 1:pos1 + 2] + "*" * (pos2 - pos1 - 2) + email[pos2:]

    return result



import inspect
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable


# initialize the context variables holding argument data
func_defaulted_args: ContextVar[dict[str, Any]] = ContextVar("defaulted_args")
func_specified_args: ContextVar[dict[str, Any]] = ContextVar("specified_args")

# initialize the context variables holding parameter data
func_defaulted_params: ContextVar[list[str]] = ContextVar("defaulted_params")
func_specified_params: ContextVar[list[str]] = ContextVar("specified_params")


def func_capture_args(func: Callable) -> Callable:
    """
    Create a decorator to identify arguments in a function which were defaulted, and which were explicitly passed.

    Introspect the call to *func* and make available two dictionaries in context variables:
        - *defaulted_args*: arguments not passed, but defaulted as per their respective declarations
        - *specified_qrgs*: arguments explicitly passed (named, positionally in *args*, or keyworded in *kwargs*)

    The need for a decorator, rather than inspecting the function at its start, results from the fact that,
    once a function is executing, no metadata about which arguments were explicitly passed vs. filled in
    by default is retained. The *inspect* module can show the current values and defaults, but not the intent
    of the caller. This is why constructs like *frame.f_locals*, *args_info*, and even *Signature.bind_partial()*
    all fall short.

    :param func: the function being decorated
    :return: the return from the call to *func*
    """
    @wraps(func)
    # ruff: noqa: ANN003 - Missing type annotation for *{name}
    def wrapper(*args, **kwargs) -> Any:

        sig = inspect.signature(func)

        # bind only explicitly passed arguments
        bound_explicit = sig.bind(*args, **kwargs)
        bound_explicit.apply_defaults()

        # bind all arguments (with defaults applied)
        bound_all = sig.bind(*args, **kwargs)
        bound_all.apply_defaults()

        # All parameters with defaults
        with_defaults = {
            name for name, param in sig.parameters.items()
            if param.default is not inspect.Parameter.empty
        }

        # explicitly passed arguments
        explicitly_passed: dict[str, Any] = dict(bound_explicit.arguments)

        # arguments that used default values
        used_defaults: dict[str, Any] = {
            name: bound_all.arguments[name]
            for name in with_defaults
            if name not in explicitly_passed
        }

        # Store in context variables
        func_specified_args.set(explicitly_passed)
        func_defaulted_args.set(used_defaults)

        # proceed executing the decorated function
        return func(*args, **kwargs)

    # prevent a rogue error ("View function mapping is overwriting an existing endpoint function")
    wrapper.__name__ = func.__name__

    return wrapper


def func_capture_params(func: Callable) -> Callable:
    """
    Create a decorator to identify parameters in a function which were defaulted, and which were explicitly passed.

    Introspect the call to *func* and make available two lists in context variables:
        - *defaulted_params*: parameters not passed, but defaulted as per their respective declarations
        - *specified_params*: parameters explicitly passed (named, positionally in *args*, or keyworded in *kwargs*)

    The need for a decorator, rather than inspecting the function at its start, results from the fact that,
    once a function is executing, no metadata about which arguments were explicitly passed vs. filled in
    by default is retained. The *inspect* module can show the current values and defaults, but not the intent
    of the caller. This is why constructs like *frame.f_locals*, *args_info*, and even *Signature.bind_partial()*
    all fall short.

    :param func: the function being decorated
    :return: the return from the call to *func*
    """
    @wraps(func)
    # ruff: noqa: ANN003 - Missing type annotation for *{name}
    def wrapper(*args, **kwargs) -> Any:

        sig = inspect.signature(func)
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        # all parameters with defaults
        with_defaults: set[str] = {
            name for name, param in sig.parameters.items()
            if param.default is not inspect.Parameter.empty
        }

        # explicitly passed parameters (from bind, before defaults applied)
        explicitly_passed: set[str] = set(sig.bind(*args, **kwargs).arguments.keys())

        # parameters that used default values
        used_defaults: set[str] = with_defaults - explicitly_passed

        # store in context variables
        func_specified_params.set(list(explicitly_passed))
        func_defaulted_params.set(list(used_defaults))

        # proceed executing the decorated function
        return func(*args, **kwargs)

    # prevent a rogue error ("View function mapping is overwriting an existing endpoint function")
    wrapper.__name__ = func.__name__

    return wrapper



from .datetime_pomes import (
    TZ_LOCAL, DateFormat, DatetimeFormat,
    date_reformat, date_weekday,
    date_parse, datetime_parse,
    timestamp_interval, timestamp_duration
)
from .dict_pomes import (
    dict_keys, dict_values, dict_items,
    dict_has_key, dict_has_value, dict_get_value, dict_set_value,
    dict_reduce, dict_listify, dict_transform, dict_merge, dict_coalesce,
    dict_clone, dict_get_key, dict_get_keys, dict_from_object, dict_from_list,
    dict_replace_value, dict_pop, dict_pop_all, dict_unique_values,
    dict_jsonify, dict_hexify, dict_stringify
)
from .email_pomes import (
    EmailParam, email_setup, email_send, email_codify,
)
from .encoding_pomes import (
    encode_ascii_hex, decode_ascii_hex
)
from .enum_pomes import (
    IntStrEnum, IntEnumUseName,
    IntEnumDesc, IntEnumCode, StrEnumDesc, StrEnumCode
)
from .env_pomes import (
    APP_PREFIX,
    env_get_str, env_get_strs,
    env_get_int, env_get_ints,
    env_get_float, env_get_floats,
    env_get_enum, env_get_enums,
    env_get_bool, env_get_bytes,
    env_get_date, env_get_path, env_get_obj, env_is_docker
)
from .file_pomes import (
    TEMP_FOLDER, Mimetype,
    file_get_data, file_get_extension,
    file_get_mimetype, file_is_binary
)
from .func_pomes import (
    func_capture_args, func_defaulted_args, func_specified_args,
    func_capture_params, func_defaulted_params, func_specified_params
)
from .list_pomes import (
    list_compare, list_correlate, list_is_slice, list_is_sub,
    list_bin_search, list_flatten, list_unflatten, list_get_coupled,
    list_elem_starting_with, list_elem_with_attr, list_transform,
    list_prune_duplicates, list_prune_in, list_prune_not_in,
    list_jsonify, list_hexify, list_hierarchize, list_stringify
)
from .obj_pomes import (
    obj_is_serializable, obj_to_dict, obj_positional, exc_format
)
from .str_pomes import (
    DIACRITICS_LOWER, DIACRITICS_UPPER,
    str_to_hex, str_from_hex, str_to_lower, str_to_upper,
    str_as_list, str_sanitize, str_split_on_mark,
    str_between, str_random, str_splice,
    str_find_char, str_find_whitespace, str_rreplace,
    str_from_any, str_to_bool, str_to_int, str_to_float,
    str_is_int, str_is_float, str_is_hex
)
from .validation_msgs import (
    validate_set_msgs, validate_update_msgs
)
from .validation_pomes import (
    VALIDATION_MSG_LANGUAGE, VALIDATION_MSG_PREFIX, MsgLang,
    validate_value, validate_bool, validate_decimal,
    validate_int, validate_ints, validate_str, validate_strs,
    validate_enum, validate_enums, validate_date, validate_datetime,
    validate_email, validate_pwd, validate_cron,
    validate_format_error, validate_format_errors, validate_unformat_errors
)
from .xml_pomes import (
    XML_FILE_HEADER,
    xml_to_dict, xml_normalize_keys
)

__all__ = [
    # __init__
    "pypomes_versions",
    # datetime_pomes
    "TZ_LOCAL", "DateFormat", "DatetimeFormat",
    "date_reformat", "date_weekday",
    "date_parse", "datetime_parse",
    "timestamp_interval", "timestamp_duration",
    # dict_pomes
    "dict_keys", "dict_values", "dict_items",
    "dict_has_key", "dict_has_value", "dict_get_value", "dict_set_value",
    "dict_reduce", "dict_listify", "dict_transform", "dict_merge", "dict_coalesce",
    "dict_clone", "dict_get_key", "dict_get_keys", "dict_from_object", "dict_from_list",
    "dict_replace_value", "dict_pop", "dict_pop_all", "dict_unique_values",
    "dict_jsonify", "dict_hexify", "dict_stringify",
    # email_pomes
    "EmailParam", "email_setup", "email_send", "email_codify",
    # encoding_pomes
    "encode_ascii_hex", "decode_ascii_hex",
    # enum_pomes
    "IntStrEnum", "IntEnumUseName",
    "IntEnumDesc", "IntEnumCode", "StrEnumDesc", "StrEnumCode",
    # env_pomes
    "APP_PREFIX",
    "env_get_str", "env_get_strs",
    "env_get_int", "env_get_ints",
    "env_get_float", "env_get_floats",
    "env_get_enum",  "env_get_enums",
    "env_get_bool", "env_get_bytes",
    "env_get_date", "env_get_path", "env_get_obj", "env_is_docker",
    # file_pomes
    "TEMP_FOLDER", "Mimetype",
    "file_get_data", "file_get_extension",
    "file_get_mimetype", "file_is_binary",
    # func_pomes
    "func_capture_args", "func_defaulted_args", "func_specified_args",
    "func_capture_params", "func_defaulted_params", "func_specified_params",
    # list_pomes
    "list_compare", "list_correlate", "list_is_slice", "list_is_sub",
    "list_bin_search", "list_flatten", "list_unflatten", "list_get_coupled",
    "list_elem_starting_with", "list_elem_with_attr", "list_transform",
    "list_prune_duplicates", "list_prune_in", "list_prune_not_in",
    "list_jsonify", "list_hexify", "list_hierarchize", "list_stringify",
    # obj_pomes
    "IntEnumUseName", "IntEnumDesc", "StrEnumDesc",
    "obj_is_serializable", "obj_to_dict", "obj_positional", "exc_format",
    # str_pomes
    "DIACRITICS_LOWER", "DIACRITICS_UPPER",
    "str_to_hex", "str_from_hex", "str_to_lower", "str_to_upper",
    "str_as_list", "str_sanitize", "str_split_on_mark",
    "str_between", "str_random", "str_splice",
    "str_find_char", "str_find_whitespace", "str_rreplace",
    "str_from_any", "str_to_bool", "str_to_int", "str_to_float",
    "str_is_int", "str_is_float", "str_is_hex",
    # validation_msgs
    "validate_set_msgs", "validate_update_msgs",
    # validation_pomes
    "VALIDATION_MSG_LANGUAGE", "VALIDATION_MSG_PREFIX", "MsgLang",
    "validate_value", "validate_bool", "validate_int", "validate_decimal",
    "validate_str", "validate_date", "validate_datetime", "validate_enum",
    "validate_email", "validate_pwd", "validate_cron",
    "validate_ints", "validate_strs", "validate_enums",
    "validate_format_error", "validate_format_errors", "validate_unformat_errors",
    # xml_pomes
    "XML_FILE_HEADER",
    "xml_to_dict", "xml_normalize_keys"
]

from contextlib import suppress
from importlib import import_module
from importlib.metadata import version
__version__: str = version("pypomes_core")
__version_info__: tuple = tuple(int(i) for i in __version__.split(".") if i.isdigit())


def pypomes_versions() -> dict[str, str]:
    """
    Retrieve the versions of the *Pypomes* packages in use.

    :return: the versions of the Pypomes packages in use
    """
    result: dict[str, str] = {
        "PyPomes-Core": __version__
    }

    with suppress(Exception):
        result["PyPomes-Cloud"] = import_module(name="pypomes_cloud").__version__

    with suppress(Exception):
        result["PyPomes-Crypto"] = import_module(name="pypomes_crypto").__version__

    with suppress(Exception):
        result["PyPomes-DB"] = import_module(name="pypomes_db").__version__

    with suppress(Exception):
        result["PyPomes-HTTP"] = import_module(name="pypomes_http").__version__

    with suppress(Exception):
        result["PyPomes-IAM"] = import_module(name="pypomes_iam").__version__

    with suppress(Exception):
        result["PyPomes-JWT"] = import_module(name="pypomes_jwt").__version__

    with suppress(Exception):
        result["PyPomes-LDAP"] = import_module(name="pypomes_ldap").__version__

    with suppress(Exception):
        result["PyPomes-Logging"] = import_module(name="pypomes_logging").__version__

    with suppress(Exception):
        result["PyPomes-Messaging"] = import_module(name="pypomes_messaging").__version__

    with suppress(Exception):
        result["PyPomes-S3"] = import_module(name="pypomes_s3").__version__

    with suppress(Exception):
        result["PyPomes-Scheduling"] = import_module(name="pypomes_scheduling").__version__

    with suppress(Exception):
        result["PyPomes-SOAP"] = import_module(name="pypomes_soap").__version__

    with suppress(Exception):
        result["PyPomes-SOB"] = import_module(name="pypomes_sob").__version__

    return result



import filetype
import puremagic
import mimetypes
from contextlib import suppress
from enum import StrEnum
from io import BytesIO, StringIO
from pathlib import Path
from tempfile import gettempdir
from typing import Final

from .env_pomes import APP_PREFIX, env_get_path

TEMP_FOLDER: Final[Path] = env_get_path(key=f"{APP_PREFIX}_TEMP_FOLDER",
                                        def_value=Path(gettempdir()))


# see https://mimetype.io/all-types
class Mimetype(StrEnum):
    """
    Commonly used mimetypes.
    """
    AAC = "audio/aac"
    AVI = " video/x-msvideo"
    BINARY = "application/octet-stream"
    BMP = "image/bmp"
    BZIP = "application/x-bzip"
    CSS = "text/css"
    CSV = "text/csv"
    DER = "application/x-x509-ca-cert"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    DWG = "image/vnd.dwg"
    EXE = "application/x-msdownload"
    FLAC = "audio/x-flac"
    FLV = "video/x-flv"
    GIF = "image/gif"
    GZIP = "application/gzip"
    HTML = "text/html"
    ICO = "image/x-icon"
    JAR = "application/java-archive"
    JAVASCRIPT = "text/javascript"
    JPEG = "image/jpeg"
    JPX = "image/jpx"
    JS = "text/javascript"
    JSON = "application/json"
    MKV = "video/x-matroska"
    MSWORD = "application/msword"
    MP3 = "audio/mpeg"
    MP4 = "video/mp4"
    MPEG = "video/mpeg"
    MIDI = "audio/midi"
    MULTIPART = "multipart/form-data"
    ODP = "application/vnd.oasis.opendocument.presentation"
    ODS = "application/vnd.oasis.opendocument.spreadsheet"
    OGG = "audio/ogg"
    P7B = "application/x-pkcs7-certificates"
    P7S = "application/pkcs7-signature"
    PDF = "application/pdf"
    PNG = "image/png"
    PPT = "application/vnd.ms-powerpoint"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    PSD = "image/vnd.adobe.photoshop"
    RAR = "application/x-rar-compressed"
    RTF = "application/rtf"
    SOAP = "application/soap+xml"
    SWF = "application/x-shockwave-flash"
    TEXT = "text/plain"
    TIFF = "image/tiff"
    URLENCODED = "application/x-www-form-urlencoded"
    WASM = "application/wasm"
    WAV = "audio/x-wav"
    WEBM = "audio/webm"
    WEBP = "image/webp"
    X7Z = "application/x-7z-compressed"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    XML = "application/xml"
    ZIP = "application/zip"


def file_get_data(file_data: BytesIO | StringIO | Path | str | bytes,
                  max_len: int = None,
                  chunk_size: int = None) -> bytes | None:
    """
    Retrieve the data in *file_data*, as implicitly defined by its data type.

    The distinction is made with the parameter's type:
        - type *BytesIO*: *file_data* is a stream of bytes (collect and returned as is)
        - type *StringIO*: *file_data* is a stream of characters (collect and returned as utf8-encoded)
        - type *Path*: *file_data* is a path to a file holding the data
        - type *str*: *file_data* holds the data (returned as utf8-encoded)
        - type *bytes*: *file_data* holds the data (returned as is)

    :param file_data: the data as implicitly defined by its data type
    :param max_len: optional maximum length of the data to return, defaults to all data
    :param chunk_size: optional chunk size to use in reading the data, defaults to 128 KB
    :return: the data, or *None* if the file data could not be obtained
    """
    # initialize the return variable
    result: bytes | None = None

    # normalize the maximum length parameter
    if isinstance(max_len, bool) or \
       not isinstance(max_len, int) or max_len < 0:
        max_len = 0

    # normalize the chunk size
    if isinstance(chunk_size, bool) or \
       not isinstance(chunk_size, int) or chunk_size <= 0:
        chunk_size = 128 * 1024

    # what is the argument type ?
    if isinstance(file_data, bytes):
        # argument is type 'bytes'
        result = file_data

    elif isinstance(file_data, str):
        # argument is type 'str'
        result = file_data.encode(encoding="utf-8")

    elif isinstance(file_data, BytesIO | StringIO):
        # argument is type 'stream'
        file_data.seek(0)
        result = file_data.read()
        if isinstance(result, str):
            result = result.encode(encoding="utf-8")

    elif isinstance(file_data, Path):
        # argument is a file path
        file_bytes: bytearray = bytearray()
        file_path: Path = Path(file_data)
        # get the data
        with file_path.open(mode="rb") as f:
            buf_size: int = min(max_len, chunk_size) if max_len else chunk_size
            in_bytes: bytes = f.read(buf_size)
            while in_bytes:
                file_bytes += in_bytes
                if max_len:
                    if max_len <= len(file_bytes):
                        break
                    buf_size = min(max_len - len(file_bytes), chunk_size)
                else:
                    buf_size = chunk_size
                in_bytes = f.read(buf_size)
        result = bytes(file_bytes)

    if result and max_len and len(result) > max_len:
        result = result[:max_len]

    return result


def file_get_extension(mimetype: Mimetype | str) -> str | None:
    """
    Obtain and return the file extension best associated with mime type *mimetype*.

    This complements Python's *mimetypes.guess_extension()*, which fails for a number of
    frequently used mime types. In those cases, if *mimetype* is an instance of *Mimetype*,
    its *enum* name is used as extension. For consistency, the extension is returned as a
    lowercase string, with the leading dot ('.').

    :param mimetype: the reference mime type
    :return: the extension, with the leading dot ('.'), best associated with *mimetype*, or *None* on fail.
    """
    # initialize the return variable
    result: str = mimetypes.guess_extension(type=mimetype)

    if not result and mimetype in Mimetype:
        mimetype = Mimetype(mimetype)
        # noinspection PyProtectedMember
        result = f".{mimetype._name_.lower()}"

    return result


def file_get_mimetype(file_data: Path | str | bytes) -> Mimetype | str:
    """
    Heuristics to determine the mimetype for *file_data*.

    The content is retrieved for analysis according to *file_data*'s type:
        - type *bytes*: *file_data* holds the data
        - type *str*: *file_data* holds the data as utf8-encoded
        - type *Path*: *file_data* is a path to a file holding the data

    The heuristics used, as heuristics go, provides an educated guess, not an accurate result.
    If a mimetype is found, and it is not in *Mimetype* (which is a small subset of known mimetypes),
    then its identifying string is returned.

    :param file_data: file data, or the path to locate the file
    :return: the probable mimetype, as a *Mimetype* object or as a string
    """
    mimetype: str | None = None
    if isinstance(file_data, str):
        file_data = file_data.encode()
    with suppress(TypeError):
        kind: filetype.Type = filetype.guess(obj=file_data)
        if kind:
            mimetype = kind.mime

    if not mimetype:
        with suppress(puremagic.PureError):
            if isinstance(file_data, Path):
                mimetype = puremagic.from_file(filename=file_data,
                                               mime=True)
            else:
                mimetype = puremagic.from_string(string=file_data,
                                                 mime=True)
    result: Mimetype | str
    if mimetype:
        # for unknown mimetypes, return its identifying string
        result = Mimetype(mimetype) if mimetype in Mimetype else mimetype
    elif file_is_binary(file_data=file_data):
        result = Mimetype.BINARY
    else:
        result = Mimetype.TEXT

    return result


def file_is_binary(file_data: Path | str | bytes) -> bool:
    """
    Heuristics to determine whether the content of *file_data* is binary.

    The content is retrieved for analysis according to *file_data*'s type:
        - type *bytes*: *file_data* holds the data
        - type *str*: *file_data* holds the data as utf8-encoded
        - type *Path*: *file_data* is a path to a file holding the data

    The heuristics used, as heuristics go, provide an educated guess, not an accurate result.
    In the present case, the first 4 KBytes of the file data is inspected for the occurrence
    of characters normally absent in text files. Thus, the presence of characters other than
    *bell*, *backspace*, *horizontal tab*, *newline*, *form feed*, *carriage return*, *escape*,
    and those in ASCII range [32 - 255] (except 127), would flag the file as binary.
    Empty or null content is considered to be non-binary.

    :param file_data: file data, or the path to locate the file
    :return: *True* if the determination resulted positive, *False* otherwise
    """
    # obtain up to 1024 bytes of content for analysis
    chunk: bytes = file_get_data(file_data=file_data,
                                 max_len=4096) or b""
    # check for null byte
    result: bool = b"\0" in chunk

    # check for non-printable characters
    if not result:
        # remove the chars listed below - chars remaining indicates content is binary
        #    7: \a (bell)
        #    8: \b (backspace)
        #    9: \t (horizontal tab)
        #   10: \n (newline)
        #   12: \f (form feed)
        #   13: \r (carriage return)
        #   27: \x1b (escape)
        #   0x20 - 0x100, less 0x7f: 32-255 char range, less 127 (the DEL control char)
        text_characters = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
        translation: bytes = chunk.translate(None,
                                             delete=text_characters)
        result = bool(translation)

    return result



from enum import IntEnum, StrEnum
from typing import Self, TypeVar

# define a TypeVar constrained to IntEnum/StrEnum
IntStrEnum = TypeVar("IntStrEnum", bound=IntEnum | StrEnum)


class IntEnumUseName(IntEnum):
    """
    A marker indicating that the attribute *name* should be used in lieu of *value*, when locating an
    instance of this class by looking for a *str* having its name. The usual procedure would be to look
    for an *int* having its value.

    Examples of such usage may be found in the *env_pomes* module. Note that this is the only situation
    justifying the use of this marker.
    """


class IntEnumDesc(IntEnum):
    """
     Adds the attribute *desc* to subclasses of *IntEnum*.
    """

    def __new__(cls,
                value: int,
                desc: str) -> Self:
        """
        Allow for the creation of subclasses of *IntEnum* having the extra *desc* attribute.

        The following are examples of creating the class *Status* with 3 instances:
        ::
            statically:
            class Status(IntEnumDesc):
                ACTIVE = (1, "Item is currently active")
                INACTIVE = (0, "Item is currently inactive")
                DELETED = (-1, "Item has been removed")

            dynamically:
            Status = IntEnumDesc("Status", {
                                 "ACTIVE": (1, "Item is currently active"),
                                 "INACTIVE": (0, "Item is currently inactive"),
                                 "DELETED": (-1, "Item has been removed")
                                 })

        :param value: value to be assigned to the instance
        :param desc: description to be assigned to the instance
        """
        result: int = int.__new__(cls,
                                  value)
        result._value_ = value
        result.desc: str = desc

        return result

    @classmethod
    def from_desc(cls,
                  value: str) -> IntEnumDesc | None:
        """
        Retrieve the instance having the attribute *desc* equal to *value*.

        :param value: the value to search for
        :return: the instance whose *desc* attributes matches *value*, or *None* if not found
        """
        # initialize the return variable
        result: IntEnumDesc | None = None

        # traverse the class
        for member in cls:
            if member.desc == value:
                result = member
                break

        return result


class IntEnumCode(IntEnum):
    """
     Adds the attributes *desc* and *code* to subclasses of *IntEnum*.
    """

    def __new__(cls,
                value: int,
                desc: str,
                code: str) -> Self:
        """
        Allow for the creation of subclasses of *IntEnum* having the extra *desc* and *code* attributes.

        The following are examples of creating the class *Status* with 3 instances:
        ::
            statically:
            class Status(IntEnumDesc):
                ACTIVE = (1, "Item is currently active", "C1")
                INACTIVE = (0, "Item is currently inactive", "C2")
                DELETED = (-1, "Item has been removed", "C3")

            dynamically:
            Status = IntEnumDesc("Status", {
                                 "ACTIVE": (1, "Item is currently active", "C1"),
                                 "INACTIVE": (0, "Item is currently inactive", "C2"),
                                 "DELETED": (-1, "Item has been removed", "C3")
                                 })

        :param value: value to be assigned to the instance
        :param desc: description to be assigned to the instance
        :param code: extra information to be assigned to the instance
        """
        result: int = int.__new__(cls,
                                  value)
        result._value_ = value
        result.desc: str = desc
        result.code: str = code

        return result

    @classmethod
    def from_code(cls,
                  value: str) -> IntEnumCode | None:
        """
        Retrieve the instance having the attribute *code* equal to *value*.

        :param value: the value to search for
        :return: the instance whose *code* attributes matches *value*, or *None* if not found
        """
        # initialize the return variable
        result: IntEnumCode | None = None

        # traverse the class
        for member in cls:
            if member.code == value:
                result = member
                break

        return result


class StrEnumDesc(StrEnum):
    """
     Adds the attribute *desc* to subclasses of *StrEnum*.
    """

    def __new__(cls,
                value: str,
                desc: str) -> Self:
        """
        Allow for the creation of subclasses of *StrEnum* having the extra *desc* attribute.

        The following are examples of creating the class *Status* with 3 instances:
        ::
            statically:
            class Status(StrEnumDesc):
                ACTIVE = ("active", "Item is currently active")
                INACTIVE = ("inactive", "Item is currently inactive")
                DELETED = ("deleted", "Item has been removed")

            dynamically:
            Status = StrEnumDesc("Status", {
                                 "ACTIVE": ("active", "Item is currently active"),
                                 "INACTIVE": ("inactive", "Item is currently inactive"),
                                 "DELETED": ("deleted", "Item has been removed")
                                 })

        :param value: value to be assigned to the instance
        :param desc: description to be assigned to the instance
        """
        result: str = str.__new__(cls,
                                  object=value)
        result._value_ = value
        result.desc: str = desc

        return result

    @classmethod
    def from_desc(cls,
                  value: str) -> StrEnumDesc | None:
        """
        Retrieve the instance having the attribute *desc* equal to *value*.

        :param value: the value to search for
        :return: the instance whose *desc* attributes matches *value*, or *None* if not found
        """
        # initialize the return variable
        result: StrEnumDesc | None = None

        # traverse the class
        for member in cls:
            if member.desc == value:
                result = member
                break

        return result


class StrEnumCode(StrEnumDesc):
    """
    Adds the attributes *desc* and *code* to subclasses of *StrEnum*.
    """

    def __new__(cls,
                value: str,
                desc: str,
                code: str) -> Self:
        """
        Allow for the creation of subclasses of *StrEnum* having the extra *desc* and *code* attributes.

        The following are examples of creating the class *Status* with 3 instances:
        ::
            statically:
            class Status(StrEnumDesc):
                ACTIVE = ("active", "Item is currently active", "C1")
                INACTIVE = ("inactive", "Item is currently inactive", "C2")
                DELETED = ("deleted", "Item has been removed", "C3")

            dynamically:
            Status = StrEnumDesc("Status", {
                                 "ACTIVE": ("active", "Item is currently active", "C1"),
                                 "INACTIVE": ("inactive", "Item is currently inactive", "C2"),
                                 "DELETED": ("deleted", "Item has been removed", "C3")
                                 })

        :param value: value to be assigned to the instance
        :param desc: description to be assigned to the instance
        :param code: extra information to be assigned to the instance
        """
        result: str = str.__new__(cls,
                                  object=value)
        result._value_ = value
        result.desc: str = desc
        result.code: str = code

        return result

    @classmethod
    def from_code(cls,
                  value: str) -> StrEnumCode | None:
        """
        Retrieve the instance having the attribute *code* equal to *value*.

        :param value: the value to search for
        :return: the instance whose *code* attributes matches *value*, or *None* if not found
        """
        # initialize the return variable
        result: StrEnumCode | None = None

        # traverse the class
        for member in cls:
            if member.code == value:
                result = member
                break

        return result
